"""
Utility to convert diaggui XML output into easily readable and introspectable HDF5 [.h5] file formats. HDF5 may be read from python using the h5py module or in matlab using the commands h5read, h5disp and similar (refer to http://www.mathworks.com/help/matlab/high-level-functions.html).

If you have hdf5 command line tools (h5ls) then you can list the datasets with

h5ls -r filename.h5
"""
import argparse
import os.path as path

import h5py

from declarative.bunch.hdf_deep_bunch import HDFDeepBunch
from .dtt2bunch import dtt2bunch


def main(args = None):
    parser = argparse.ArgumentParser(
        prog = 'dtt2hdf5',
        description=__doc__,
    )

    parser.add_argument(
        #'--control-A',
        dest     = 'file_from',
        metavar  = 'from-file',
        type     = str,
        help     = 'diaggui xml file to parse'
    )

    parser.add_argument(
        #'--control-A',
        dest     = 'file_to',
        metavar  = 'to-file',
        type     = str,
        nargs    = '?',
        default  = None,
        help     = 'HDF5 file to insert to. Will rename the input to a .h5 if not given.'
    )

    parser.add_argument(
        '-q, --quiet',
        dest     = 'quiet',
        action   = 'store_true',
        help     = 'Quiet the output',
    )

    args = parser.parse_args(args = args)

    ffrom_base, ffrom_ext = path.splitext(args.file_from)
    if args.file_to is None:
        file_to = ffrom_base + '.h5'
    else:
        file_to = args.file_to

    if not args.quiet:
        print(file_to)

    #parser.add_argument(
    #    '-E',
    #    dest='control_A',
    #    type=int,
    #    required=True,
    #    help='Exclude Channels'
    #)

    diag_DB = dtt2bunch(fpath = args.file_from, verbose = not args.quiet)

    #write the file out, it is cleared first due to the 'w' flag. Could do this using append operations with 'a'
    with h5py.File(file_to, 'w') as F_hdf:
        #TODO, using _vpath is deprecated in declarative > 1.2.0
        hdf = HDFDeepBunch(F_hdf, writeable = True, _vpath = True)
        hdf.update_recursive(diag_DB)


if __name__ == '__main__':
    main()



"""
"""
import numpy as np
from .parse import dtt_read


class DiagMeasurementHolder(object):
    _copy_elements = ('gps_second', 'window', 'averages', 'BW')

    gps_second  = None
    window      = None
    averages    = None
    BW          = None
    diag_access = None

    def _metadata_mirror(self, other):
        for attr_name in self._copy_elements:
            setattr(self, attr_name, getattr(other, attr_name))
        return

    def metadata_check(self, other):
        for attr_name in self._copy_elements:
            a = getattr(self, attr_name)
            b = getattr(other, attr_name)
            if a != b:
                raise RuntimeError(
                    ("Metadata field {0} doesn't match ({1} != {2})"
                     ).format(attr_name, a, b)
                )
        return

    @property
    def utc_time(self):
        """
        Add the GPS to UTC epoch time difference, with the leap seconds included
        """
        return self.gps_second + 315964784


class DiagFreqMeasurementHolder(DiagMeasurementHolder):
    def _metadata_mirror(self, other):
        super(DiagFreqMeasurementHolder, self)._metadata_mirror(other)
        self.FHz = other.FHz
        return

    def metadata_check(self, other):
        super(DiagFreqMeasurementHolder, self).metadata_check(other)
        if not all(other.FHz == self.FHz):
            raise RuntimeError("Freqs supposed to be the same")
        return


class DiagCoherenceHolder(DiagFreqMeasurementHolder):
    def __init__(self, daccess, chn1, chn2):
        self.daccess = daccess
        self.chn1 = chn1
        self.chn2 = chn2
        try:
            chn1_coh = self.daccess.results['COH'][chn1]
            self._metadata_mirror(chn1_coh)
            idx = chn1_coh.channelB_inv[chn2]
            self.coh = chn1_coh.coherence[idx].reshape(-1)
        except KeyError:
            chn2_coh = self.daccess.results['COH'][chn2]
            self._metadata_mirror(chn2_coh)
            idx = chn2_coh.channelB_inv[chn1]
            self.coh = chn2_coh.coherence[idx].reshape(-1)

    _coh_phased = None
    @property
    def coh_phased(self):
        if self._coh_phased is not None:
            return self._coh_phased
        m_csd = self.daccess.csd(self.chn1, self.chn2)
        self.metadata_check(m_csd)
        self._coh_phased = self.coh * m_csd.csd / abs(m_csd.csd)
        return self._coh_phased

    @property
    def statistical_minimum(self):
        return 1/self.averages**.5


class DiagCSDHolder(DiagFreqMeasurementHolder):
    def __init__(self, daccess, chn1, chn2):
        self.daccess = daccess
        self.chn1 = chn1
        self.chn2 = chn2
        try:
            chn1_csd = self.daccess.results['CSD'][chn1]
            self._metadata_mirror(chn1_csd)
            idx = chn1_csd.channelB_inv[chn2]
            self.csd = chn1_csd.CSD[idx].reshape(-1)
        except KeyError:
            chn2_csd = self.daccess.results['CSD'][chn2]
            self._metadata_mirror(chn2_csd)
            idx = chn2_csd.channelB_inv[chn1]
            self.csd = chn2_csd.CSD[idx].reshape(-1)


class DiagASDHolder(DiagFreqMeasurementHolder):
    def __init__(self, daccess, chn):
        self.daccess = daccess
        self.chn = chn
        chn_asd = self.daccess.results['PSD'][chn]
        self._metadata_mirror(chn_asd)
        self.asd = chn_asd.PSD.reshape(-1)


class DiagXferHolder(DiagFreqMeasurementHolder):
    def __init__(self, daccess, chn_num, chn_den):
        self.daccess = daccess
        self.chn_num = chn_num
        self.chn_den = chn_den

        self.xfer = None
        try:
            self.chn_den_TF = self.daccess.results['TF'][self.chn_den]
            self._metadata_mirror(self.chn_den_TF)
            idx = self.chn_den_TF.channelB_inv[self.chn_num]
            self.xfer = self.chn_den_TF.xfer[idx].reshape(-1)
        except KeyError:
            pass

        if self.xfer is None:
            try:
                self.chn_num_TF = self.daccess.results['TF'][self.chn_num]
                self._metadata_mirror(self.chn_num_TF)
                idx = self.chn_num_TF.channelB_inv[self.chn_den]
                self.xfer = 1./self.chn_num_TF.xfer[idx].reshape(-1)
            except KeyError:
                pass

        #if the measurment aborts before it completes, it pads the freq data with zeros
        #if self.xfer is not None:
            #freq_good_idx = self.FHz > 0.
            #self.xfer = self.xfer[freq_good_idx]
            #self.FHz = self.FHz[freq_good_idx]

        if self.xfer is None:
            try:
                self._metadata_mirror(self.csd_obj)
                self.metadata_check(self.asd_den_obj)
                self.xfer = self.csd_obj.csd / self.asd_den_obj.asd**2
            except KeyError:
                raise
                raise RuntimeError("Diag file does not seem to carry this xfer function")

    _coh_obj = None
    @property
    def coh_obj(self):
        if self._coh_obj is not None:
            return self._coh_obj
        self._coh_obj = self.daccess.coherence(self.chn_num, self.chn_den)
        self.metadata_check(self._coh_obj)
        return self._coh_obj

    @property
    def coh(self):
        return self.coh_obj.coh

    @property
    def SNR_estimate(self):
        return (self.coh/np.maximum(1/self.averages, 1 - self.coh - 1/self.averages) * self.averages)**0.5

    _csd_obj = None
    @property
    def csd_obj(self):
        if self._csd_obj is None:
            self._csd_obj = self.daccess.csd(self.chn_num, self.chn_den)
        return self._csd_obj

    @property
    def csd(self):
        return self.csd_obj.csd

    _asd_num_obj = None
    @property
    def asd_num_obj(self):
        if self._asd_num_obj is not None:
            return self._asd_num_obj
        self._asd_num_obj = self.daccess.asd(self.chn_num)
        self.metadata_check(self._asd_num_obj)
        return self._asd_num_obj

    @property
    def asd_num(self):
        return self.asd_num_obj.asd

    _asd_den_obj = None
    @property
    def asd_den_obj(self):
        if self._asd_num_obj is None:
            self._asd_den_obj = self.daccess.asd(self.chn_den)
        return self._asd_den_obj

    @property
    def asd_den(self):
        return self._asd_den_obj.asd


class DiagXferViaHolder(DiagFreqMeasurementHolder):

    def __init__(self, daccess, chn_num, chn_den, chn_via):
        self.daccess = daccess
        self._xfer_num_obj = self.daccess.xfer(chn_num, chn_via)
        self._xfer_den_obj = self.daccess.xfer(chn_den, chn_via)
        self._metadata_mirror(self._xfer_num_obj)
        self.metadata_check(self._xfer_den_obj)
        self.xfer = self._xfer_num_obj.xfer / self._xfer_den_obj.xfer

    @property
    def SNR_estimate(self):
        snr_num = self._xfer_num_obj.SNR_estimate
        snr_den = self._xfer_den_obj.SNR_estimate
        return (snr_den**-2 + snr_num**-2)**(-0.5)

    _coh_nv_obj = None
    @property
    def coh_nv_obj(self):
        if self._coh_nv_obj is not None:
            return self._coh_nv_obj
        self._coh_nv_obj = self.daccess.coherence(self.chn1, self.chn2)
        self.metadata_check(self._coh_nv_obj)
        return self._coh_nv_obj

    @property
    def coh_nv(self):
        return self._coh_nv_obj.coh

    _coh_dv_obj = None
    @property
    def coh_dv_obj(self):
        if self._coh_dv_obj is not None:
            return self._coh_dv_obj
        self._coh_dv_obj = self.daccess.coherence(self.chn1, self.chn2)
        self.metadata_check(self._coh_dv_obj)
        return self._coh_dv_obj

    @property
    def coh_dv(self):
        return self._coh_dv_obj.coh

    _coh_nd_obj = None
    @property
    def coh_nd_obj(self):
        if self._coh_nd_obj is not None:
            return self._coh_nd_obj
        self._coh_nd_obj = self.daccess.coherence(self.chn1, self.chn2)
        self.metadata_check(self._coh_nd_obj)
        return self._coh_nd_obj

    @property
    def coh_nd(self):
        return self._coh_nd_obj.coh

    @property
    def asd_num_obj(self):
        return self._xfer_num_obj.asd_num_obj

    @property
    def asd_via_obj(self):
        return self._xfer_num_obj.asd_den_obj

    @property
    def asd_den_obj(self):
        return self._xfer_den_obj.asd_num_obj

    @property
    def asd_num(self):
        return self.asd_num_obj.asd

    @property
    def asd_den(self):
        return self.asd_den_obj.asd

    @property
    def asd_via(self):
        return self.asd_via_obj.asd


class DiagSineResponseCoefficients(DiagMeasurementHolder):
    def __init__(self, daccess, chn_list, chn_wrt = None, freq_idx = 0):
        self.daccess = daccess
        self.chn_list = np.asarray(chn_list)
        self.chn_map = {}
        for idx, chn in enumerate(self.chn_list):
            self.chn_map[idx] = chn
            self.chn_map[chn] = idx
        self.chn_wrt = chn_wrt

        diag_TC = self.daccess.results['TransferCoefficients']
        diag_CC = self.daccess.results['CoherenceCoefficients']
        self._metadata_mirror(diag_TC)
        self.metadata_check(diag_CC)
        self.FHz = diag_TC.FHz[freq_idx]
        if not np.all(self.FHz == diag_CC.FHz[freq_idx]):
            raise RuntimeError("Tranfer and Coherence coefficient frequencies inconsistent")

        if chn_wrt is not None:
            diag_idx_wrt = diag_TC.channels_inv[chn_wrt]
            self.coeff_wrt = diag_TC.coeffs[freq_idx, diag_idx_wrt]
            self.coh_wrt = diag_CC.coeffs[freq_idx, diag_idx_wrt]
        else:
            self.coeff_wrt = 1.
            self.coh_wrt   = 1.

        self.coeffs = np.empty(len(self.chn_list), dtype = complex)
        self.coeffs_dict = {}
        self.cohs = np.empty(len(self.chn_list), dtype = float)
        self.cohs_dict = {}
        for idx, chn in enumerate(self.chn_list):
            diag_idx = diag_TC.channels_inv[chn]
            coeff = diag_TC.coeffs[freq_idx, diag_idx] / self.coeff_wrt
            self.coeffs[idx] = coeff
            self.coeffs_dict[chn] = coeff
            diag_idx = diag_CC.channels_inv[chn]
            coh = diag_CC.coeffs[freq_idx, diag_idx]
            self.cohs[idx] = coh
            self.cohs_dict[chn] = coh


class DiagHarmonicResponseCoefficients(DiagMeasurementHolder):
    def __init__(self, daccess, chn_list, chn_wrt = None, chn_wrt_freq = 1):
        self.daccess = daccess
        self.chn_list = np.asarray(chn_list)
        self.chn_map = {}
        for idx, chn in enumerate(self.chn_list):
            self.chn_map[idx] = chn
            self.chn_map[chn] = idx
        self.chn_wrt = chn_wrt
        self.chn_wrt_freq = chn_wrt_freq

        diag_HC = self.daccess.results['HarmonicCoefficients']
        self._metadata_mirror(diag_HC)
        self.FHz = diag_HC.FHz

        if chn_wrt is not None:
            diag_idx_wrt = diag_HC.channels_inv[chn_wrt]
            self.coeff_wrt = diag_HC.coeffs[chn_wrt_freq, diag_idx_wrt]
        else:
            self.coeff_wrt = 1.
            self.coh_wrt   = 1.
        fs = len(self.FHz)
        l = len(self.chn_list)
        self.coeffs = np.empty((l, fs), dtype = complex)
        self.coeffs_dict = {}
        self.coeffs_2x = np.empty(l, dtype = complex)
        self.coeffs_2x_dict = {}
        for idx, chn in enumerate(self.chn_list):
            diag_idx = diag_HC.channels_inv[chn]
            coeff = diag_HC.coeffs[:, diag_idx] / self.coeff_wrt
            self.coeffs[idx] = coeff
            self.coeffs_dict[chn] = coeff
            coeff_2x = coeff[2] / coeff[1]
            self.coeffs_2x[idx] = coeff_2x
            self.coeffs_2x_dict[chn] = coeff_2x


class DiagAccess(object):
    """High level access to measurements from diaggui

    args:
        fname (string): xml file path to read

    This class reads diaggui's .xml files and exposes an interface to wrap
    various measurements into objects. It currently cannot extract reference
    measurements. Use the low level :func:`dtt_read` function to access
    references.
    """

    def __init__(
            self,
            fname,
    ):
        # Parse the file
        raw = dtt_read(fname)
        self.references = raw.references
        self.results    = raw.results

        # Add the reference traces to the externally-accessible results dict
        # Channel naming convention follows the DTT convention: "CHN_NAME(REF#)"
        for idx_ref, ref in self.references.items():
            type_ref = ref['type_name']
            if type_ref in ('PSD', 'CSD', 'COH', 'TF'):
                chn_ref = '{}(REF{})'.format(ref['channelA'], idx_ref)
                if type_ref not in self.results:
                    self.results[type_ref] = dict()
                self.results[type_ref][chn_ref] = ref

    def coherence(self, chn1, chn2):
        """
        Return a :class:`DiagCoherenceHolder` containing a coherence measurement
        between two channels

        args:
            chn1 (string):
            chn2 (string):
        """
        return DiagCoherenceHolder(self, chn1, chn2)

    def coh(self, chn1, chn2):
        """
        Return a :class:`DiagCoherenceHolder` containing a coherence measurement
        between two channels

        args:
            chn1 (string):
            chn2 (string):
        """
        return DiagCoherenceHolder(self, chn1, chn2)

    def csd(self, chn1, chn2):
        """
        Return a :class:`DiagCoherenceHolder` containing a coherence measurement
        between two channels

        args:
            chn1 (string): primary channel
            chn2 (string): conjugated channel
        """
        return DiagCSDHolder(self, chn1, chn2)

    def asd(self, chn):
        """
        Return a :class:`DiagASDHolder` containing an ASD measurement for a
        single channel

        args:
            chn (string): channel
        """
        return DiagASDHolder(self, chn)

    def xfer(self, chn_num, chn_den):
        """
        Return a :class:`DiagXferHolder` containing a transfer function

        args:
            chn_num (string): numerator channel
            chn_den (string): denominator channel

        Note:
            For 'Fourier Tools' measurements, this method is biased if
            computed on two noisy channels, use :meth:`xfer_via` for unbiased
            measurements between two channels with an excitation.
        """
        return DiagXferHolder(self, chn_num, chn_den)

    def xfer_via(self, chn_num, chn_den, chn_via):
        """
        Return a :class:`DiagXferViaHolder` containing an ASD measurement for a
        single channel

        args:
            chn_num (string): numerator channel
            chn_den (string): denominator channel
            chn_via (string): phase reference channel, usually the excitation
        """
        return DiagXferViaHolder(self, chn_num, chn_den, chn_via)

    def sine_response(self, chn_list, chn_wrt = None, freq_idx = 0):
        """
        Returns a :class:`DiagSineResponseCoefficients` objects to access sine
        response measurements

        args:
            chn_list (list of str): channels to measure
            chn_wrt (string): "with respect to" reference channel, usually an excitation.
            freq_idx (int): which of the excitation frequencies to use
        """

        return DiagSineResponseCoefficients(
            self,
            chn_list = chn_list,
            chn_wrt = chn_wrt,
            freq_idx = freq_idx
        )

    def harmonic_response(self, chn_list, chn_wrt = None):
        """
        Returns a :class:`DiagHarmonicResponseCoefficients` objects to access sine
        response measurements

        args:
            chn_list (list of str): channels to measure
            chn_wrt (string): "with respect to" reference channel, usually an excitation.
        """
        return DiagHarmonicResponseCoefficients(
            self,
            chn_list = chn_list,
            chn_wrt = chn_wrt,
        )

    def channels(self):
        """
        returns two lists of channels, the Achannels and the Bchannels.
        """
        channels_A = set()
        channels_B = set()
        for res_key, res_obj in self.results.items():
            if res_key in (
                'HarmonicCoefficients',
                'CoherenceCoefficients',
                'TransferCoefficients',
                'TransferMatrix',
            ):
                channels_A.update(list(res_obj.channels.values()))
            if res_key in ('TS',):
                channels_A.update(list(res_obj.keys()))
            else:
                for chn_key, chn_obj in res_obj.items():
                    channels_A.add(chn_key)
                    channels_B.update(list(chn_obj.channelB_inv.keys()))
        #reduce the B channels to being the unique ones
        channels_B = channels_B - channels_A
        return channels_A, channels_B

    def channels_print(self):
        """
        Prints the A and B channels to stdout
        """
        channels_A, channels_B = self.channels()
        channels_A = sorted(channels_A)
        print("A-Channels:")
        for chn in channels_A:
            print("\t", chn)
        if channels_B:
            channels_B = sorted(channels_B)
            print("Non-A, B-Channels:")
            for chn in channels_B:
                print("\t", chn)





import pytest
import sys
import subprocess
from dttxml.dtt2bunch import dtt2bunch
from dttxml.dtt2hdf import main


def run_dtt2hdf(*args):
    try:
        main(args)
    except Exception:
        return False
    return True



@pytest.mark.xfail(reason = """
This dttxml has references and the bunch exporter currently cannot handle this.
It uses DiagAccess to combinatorically export everything, and the references are
not available through the DiagAccess interfaces
""")
def test_dtt2hdf_1(fpath_join, tpath_join, plot, pprint):
    """
    """
    # EXAMPLE 1: Dealing with a SweptSine measurement file.

    # These are the xml files we want to get our data from.
    measurement_file = fpath_join('data', '2020-01-03_H1_DARM_OLGTF_LF_SS_5to1100Hz_15min.xml')
    b = dtt2bunch(measurement_file)
    success = run_dtt2hdf(measurement_file, tpath_join('output.h5'))

    assert(success)


@pytest.mark.xfail(reason = """
This dttxml is internally parsed from 'transfer function B/A in format (Y)'.
Currently the parse_transfer function does not generate the FHz array using the
available data. It probably could, but needs testing against exported data.
""")
def test_dtt2hdf_2(fpath_join, tpath_join, plot, pprint):
    """
    """
    # EXAMPLE 1: Dealing with a SweptSine measurement file.

    # These are the xml files we want to get our data from.
    measurement_file = fpath_join('data', '2019-01-25_H1SUSETMX_L1_iEXC2DARM_HFDynamicsTest_40-110Hz.xml')

    b = dtt2bunch(measurement_file)
    success = run_dtt2hdf(measurement_file, tpath_join('output.h5'))

    assert(success)




#@pytest.mark.xfail(reason = """
#""")
def test_dtt2hdf_3(fpath_join, tpath_join, plot, pprint):
    """
    """
    # EXAMPLE 1: Dealing with a SweptSine measurement file.

    # These are the xml files we want to get our data from.
    measurement_file = fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.xml')

    b = dtt2bunch(measurement_file)
    pprint("----------------------output dict---------------------------")
    pprint(b)
    pprint("----------------------output done---------------------------")
    success = run_dtt2hdf(measurement_file, tpath_join('output.h5'))

    assert(success)



import numpy as np
from matplotlib import pyplot as plt

import pydarm_measurement as measurement


def test_example_1(fpath_join, tpath_join, plot, pprint):
    """
    Example code on how to use the following methods from measurement.py:
        - get_raw_tf
        - get_set_of_channels
        - get_raw_asd
    """
    # EXAMPLE 1: Dealing with a SweptSine measurement file.

    # These are the xml files we want to get our data from.
    measurement_file_1 = fpath_join('data', '2020-01-03_H1_DARM_OLGTF_LF_SS_5to1100Hz_15min.xml')

    meas_obj_1 = measurement.Measurement(measurement_file_1)
    channelA_1 = 'H1:LSC-DARM1_IN2'
    channelB_1 = 'H1:LSC-DARM1_IN1'
    temp_freq, temp_tf, temp_coh, temp_unc = meas_obj_1.get_raw_tf(channelA_1, channelB_1)

    # print(temp_raw_tf)

    plt.figure(1)
    plt.subplot(211)
    plt.loglog(temp_freq, np.abs(temp_tf), 'ro')

    plt.subplot(212)
    plt.semilogx(temp_freq, np.angle(temp_tf, deg=True), 'ro')
    if plot:
        plt.savefig(tpath_join('DARM.pdf'), format='pdf')
    #plt.show(block=False)

    # Examples of other information you can get from the measurement class:
    pprint('----------------------------------------------------')
    pprint('Other information in file ', measurement_file_1)
    pprint('GPS of measurement is ', meas_obj_1.gps_time)
    pprint('Navg of measurement is ', meas_obj_1.averages)
    pprint('Channels in measurement file are:', meas_obj_1.get_set_of_channels())
    pprint('----------------------------------------------------')


def test_example_2(fpath_join, tpath_join, plot, pprint):
    # ----------------------------------------------------------------
    # ----------------------------------------------------------------

    # EXAMPLE 2: Dealing with an FFT dtt file.

    # The following lines show an example of ASD extraction from a dtt
    # file that contains FFT (or BroadBand data)

    # These are the xml files we want to get our data from.
    measurement_file_2 = fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.xml')

    meas_obj_2 = measurement.Measurement(measurement_file_2)
    channelA_2 = 'H1:LSC-DARM1_IN2'
    channelB_2 = 'H1:LSC-DARM1_IN1'
    channelA_freq, channelA_asd = meas_obj_2.get_raw_asd(channelA_2)
    channelB_freq, channelB_asd = meas_obj_2.get_raw_asd(channelB_2)

    # print(channelA_asd)

    plt.figure(2)
    plt.loglog(channelA_freq, channelA_asd, 'k', channelB_freq, channelB_asd, 'r')
    if plot:
        plt.savefig(tpath_join('DARM.pdf'))
    #plt.show(block=True)
    
    pprint('----------------------------------------------------')
    pprint('Other information in measurement file: ', measurement_file_2)
    pprint('GPS of measurement is ', meas_obj_2.gps_time)
    pprint('Navg of measurement is ', meas_obj_2.averages)
    pprint('Channels in measurement file are:', meas_obj_2.get_set_of_channels())
    pprint('----------------------------------------------------')
    # ----------------------------------------------------------------------



# -*- coding: utf-8 -*-
from __future__ import division, print_function, unicode_literals
import os
import sys
from os import path
from distutils.command.bdist import bdist
from distutils.command.sdist import sdist

def version_checker(version_string, module_name):

    def check_versions():
        import importlib
        print('versions checked')
        try:
            curpath = path.abspath(path.realpath(os.getcwd()))
            setuppath = path.abspath(path.realpath(path.split(__file__)[0]))

            if curpath != setuppath:
                sys.path.append(setuppath)

            module = importlib.import_module(module_name)

            modfile = path.abspath(path.realpath(path.split(module.__file__)[0]))
            mod_relpath = path.relpath(modfile, setuppath)
            if mod_relpath == module_name:
                if module.__version__ != version_string:
                    print("WARNING: Stated module version different than setup.py version", file = sys.stderr)
                    print("         '{0}' != '{1}'".format(module.__version__, version_string), file = sys.stderr)
                    print("Fix version.py and setup.py for consistency")

            import subprocess
            try:
                git_tag = subprocess.check_output(['git', 'describe', '--tags'])
                git_tag = git_tag.strip()
                git_tag = str(git_tag.decode('utf-8'))
            except subprocess.CalledProcessError:
                pass
            else:
                if git_tag != version_string:
                    if git_tag.startswith(version_string):
                        print("WARNING: latex git-tag has commits since versioning", file = sys.stderr)
                        print("         '{0}' ---> '{1}'".format(version_string, git_tag), file=sys.stderr)
                    else:
                        print("WARNING: latex git-tag different than setup.py version", file = sys.stderr)
                        print("         '{0}' != '{1}'".format(version_string, git_tag), file=sys.stderr)
                    print("         Perhaps update versions in setup.py and module.version, git commit, then git tag")
                    print("         otherwise fix tag if not yet git pushed to remote (see DISTRIBUTION-README.md)")

        except ImportError:
            pass

    cmdclass = dict()

    class check_bdist(bdist):
        def run(self):
            bdist.run(self)
            check_versions()

    class check_sdist(sdist):
        def run(self):
            sdist.run(self)
            check_versions()


    cmdclass['bdist'] = check_bdist
    cmdclass['sdist'] = check_sdist

    try:
        from wheel.bdist_wheel import bdist_wheel
        class check_bdist_wheel(bdist_wheel):
            def run(self):
                bdist_wheel.run(self)
                check_versions()
    except ImportError:
        pass
    else:
        cmdclass['bdist_wheel'] = check_bdist_wheel

    return cmdclass





import base64
import numpy as np
from declarative.bunch import Bunch
import re
#import xml.etree.cElementTree as etree


def parse_spectrum(LW_node):
    specbunch = Bunch()
    specbunch.channelB = Bunch()
    specbunch.channelB_inv = Bunch()

    for sparam in LW_node.findall('Time'):
        if sparam.attrib["Name"] == 't0':
            specbunch.gps_second = float(sparam.text)
    for sparam in LW_node.findall('Param'):
        spname = sparam.attrib["Name"]
        if spname == 'Subtype':
            subtype_raw = int(sparam.text)
        elif spname == 'Window':
            window_raw = int(sparam.text)
            specbunch.window = {
                0 : 'Uniform',
                1 : 'Hanning',
                2 : 'Flat-top',
                3 : 'Welch',
                4 : 'Bartlett',
                5 : 'BMH',
                6 : 'Hamming',
                7 : 'Kaiser',
            }.get(window_raw, "unknown")
        elif spname == 'AverageType':
            avgtype_raw = int(sparam.text)
            specbunch.avgtype = {
                0 : 'Fixed',
                1 : 'Exponential',
                2 : 'Single',
            }.get(avgtype_raw, "unknown")
        elif spname == 'Averages':
            specbunch.averages = int(sparam.text)
        elif spname == 'M':
            M = int(sparam.text)
        elif spname == 'N':
            N = int(sparam.text)
        elif spname == 'f0':
            specbunch.f0 = float(sparam.text)
        elif spname == 'BW':
            specbunch.BW = float(sparam.text)
        elif spname == 'df':
            specbunch.df = float(sparam.text)
        elif spname == 'ChannelA':
            specbunch.channelA = sparam.text
        elif spname.startswith('ChannelB'):
            RefNMatch = re.match(r'ChannelB\[(\d+)\]', spname)
            chn_num = int(RefNMatch.group(1))
            specbunch.channelB[chn_num] = sparam.text
        else:
            pass

    chanB = specbunch.channelB
    sarg = np.argsort(list(chanB.keys()))
    if len(sarg) > 0:
        specbunch.channelB = np.take(list(chanB.values()), sarg)
        for idx, chn_name in enumerate(specbunch.channelB):
            specbunch.channelB_inv[chn_name] = idx

    ####
    ##Now interpret the data
    streambuff = base64.b64decode(LW_node.findall('Array/Stream')[0].text)

    if subtype_raw == 0:
        specbunch.subtype = 'FFT in format (Y)'
        data = np.frombuffer(streambuff, dtype='c8')
        specbunch.FHz = np.linspace(specbunch.f0, specbunch.f0 + specbunch.df*(N-1), N)
        specbunch.type_name = 'FFT'
        specbunch.FFT = data.reshape(M, -1)
    elif subtype_raw == 1:
        specbunch.subtype = 'power spectral density in format (Y)'
        specbunch.type_name = 'PSD'
        data = np.frombuffer(streambuff, dtype='f4')
        specbunch.FHz = np.linspace(specbunch.f0, specbunch.f0 + specbunch.df*(N-1), N)
        specbunch.PSD = data.reshape(M, -1)
    elif subtype_raw == 2:
        specbunch.subtype = 'cross-power spectrum in format (Y)'
        specbunch.type_name = 'CSD'
        data = np.frombuffer(streambuff, dtype='c8')
        specbunch.FHz = np.linspace(specbunch.f0, specbunch.f0 + specbunch.df*(N-1), N)
        specbunch.CSD = data.reshape(M, -1)
    elif subtype_raw == 3:
        specbunch.subtype = 'coherence in format (Y)'
        specbunch.type_name = 'COH'
        data = np.frombuffer(streambuff, dtype='f4')
        specbunch.FHz = np.linspace(specbunch.f0, specbunch.f0 + specbunch.df*(N-1), N)
        specbunch.coherence = data.reshape(M, -1)
    elif subtype_raw == 4:
        specbunch.subtype = 'FFT in format (f, Y)'
        specbunch.type_name = 'FFT'
        data = np.frombuffer(streambuff, dtype='c8')
        specbunch.FHz = data[:N]
        specbunch.FFT = data[N:].reshape(M, -1)
    elif subtype_raw == 5:
        specbunch.subtype = 'power spectral density in format (f,Y)'
        specbunch.type_name = 'PSD'
        data = np.frombuffer(streambuff, dtype='f4')
        specbunch.FHz = data[:N]
        specbunch.PSD = data[N:].reshape(M, -1)
    elif subtype_raw == 6:
        specbunch.subtype = 'cross-power spectrum in format (f,Y)'
        specbunch.type_name = 'CSD'
        data = np.frombuffer(streambuff, dtype='c8')
        specbunch.FHz = data[:N].real
        specbunch.CSD = data[N:].reshape(M, -1)
    elif subtype_raw == 7:
        specbunch.subtype = 'coherence in format (f,Y)'
        specbunch.type_name = 'COH'
        data = np.frombuffer(streambuff, dtype='f4')
        specbunch.FHz = data[:N]
        specbunch.coherence = data[N:].reshape(M, -1)
    else:
        specbunch.subtype = "unknown"
    return specbunch




import dttxml

import numpy as np
import matplotlib.pyplot as plt

def test_frequency_array1(fpath_join, tpath_join, plot):
    meas_file = fpath_join('data', 'test_DTT_20_f_pts.xml')

    channelA = 'H1:PEM-EX_SEIS_VEA_FLOOR_Z_DQ'
    # channelA = 'H1:CAL-PCALX_TX_PD_OUT_DQ'
    data_access = dttxml.DiagAccess(meas_file)

    asd_holder = data_access.asd(channelA)

    freq = np.array(asd_holder.FHz)
    asd = np.array(asd_holder.asd)

    data_from_dtt = np.genfromtxt(
        fpath_join('data', 'test_DTT_20_f_pts.txt'),
        dtype='float',
        delimiter=None
    )

    plt.figure(1)
    # plt.loglog(freq,asd,'r',data_from_dtt[:,0],data_from_dtt[:,1]*2,'k')
    plt.plot(freq, (data_from_dtt[:, 0]-freq), 'r')
    if plot:
        plt.savefig(tpath_join('residual1.pdf'))


def test_frequency_array2(fpath_join, tpath_join, plot, pprint):
    meas_file_2 = fpath_join('data', 'test_DTT_21_f_pts.xml')
    channelA_2 = 'H1:PEM-EX_SEIS_VEA_FLOOR_Z_DQ'

    data_access_2 = dttxml.DiagAccess(meas_file_2)
    asd_holder_2 = data_access_2.asd(channelA_2)

    freq_2 = asd_holder_2.FHz

    data_from_dtt_2 = np.genfromtxt(
        fpath_join('data', 'test_DTT_21_f_pts.txt'),
        dtype='float',
        delimiter=None
    )

    for i in range(len(freq_2)):
        pprint(freq_2[i], data_from_dtt_2[i, 0])

    plt.figure(2)
    plt.plot(freq_2, np.abs(data_from_dtt_2[:, 0]-freq_2))
    if plot:
        plt.savefig(tpath_join('residual2.pdf'))




"""
Utility to convert diaggui XML output into easily readable and introspectable HDF5 [.h5] file formats. HDF5 may be read from python using the h5py module or in matlab using the commands h5read, h5disp and similar (refer to http://www.mathworks.com/help/matlab/high-level-functions.html).

If you have hdf5 command line tools (h5ls) then you can list the datasets with

h5ls -r filename.h5
"""
from . import dtt2hdf

if __name__ == '__main__':
    dtt2hdf.main()



import numpy as np
from matplotlib import pyplot as plt
import dttxml
from numpy.testing import assert_allclose


def test_access_TF(fpath_join, tpath_join, plot, pprint):
    """
    Example code on how to use the following methods from measurement.py:
        - get_raw_tf
        - get_set_of_channels
        - get_raw_asd
    """
    # EXAMPLE 1: Dealing with a SweptSine measurement file.

    # These are the xml files we want to get our data from.
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2020-01-03_H1_'
                   'DARM_OLGTF_LF_SS_A_DARMIN2_B_DARMIN1_tf.txt'),
        dtype='float',
        delimiter=None
    )

    freq_from_dtt = data_from_dtt[:, 0]
    tf_from_dtt = data_from_dtt[:, 1]+1j*data_from_dtt[:, 2]

    measurement_file = fpath_join('data', '2020-01-03_H1_'
                                  'DARM_OLGTF_LF_SS_5to1100Hz_15min.xml')
    channelA = 'H1:LSC-DARM1_IN2'
    channelB = 'H1:LSC-DARM1_IN1'
    acc = dttxml.DiagAccess(measurement_file)
    tf = acc.xfer(channelB, channelA)

    # Note: For some reason, the txt file is missing one frequency.
    # This is unrelated to the coherence, at least of this TF.

    #TODO, dtt exports one less value than the xml file has available?!
    max_num = 22
    assert_allclose(freq_from_dtt[:max_num], tf.FHz[:max_num], atol=1e-4)
    assert_allclose(tf_from_dtt[:max_num], tf.xfer[:max_num], atol=1e-4)


def test_access_CSD(fpath_join, tpath_join, plot, pprint):
    """
    This is currently the most legit test
    """
    measurement_file = fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.xml')
    acc = dttxml.DiagAccess(measurement_file)

    #first do an ASD
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.txt'),
        dtype='float',
        delimiter=None
    )

    freq_from_dtt = data_from_dtt[:, 0]
    asd_from_dtt = data_from_dtt[:, 1]

    channelA = 'H1:LSC-DARM1_IN2'
    asd = acc.asd(channelA)

    assert_allclose(asd_from_dtt, asd.asd)
    assert_allclose(freq_from_dtt, asd.FHz)

    #now do COH
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB_COH_E1_E2_12.txt'),
        dtype='float',
        delimiter=None
    )

    freq_from_dtt = data_from_dtt[:, 0]
    COH_E1_from_dtt = data_from_dtt[:, 1]
    COH_E2_from_dtt = data_from_dtt[:, 2]
    COH_12_from_dtt = data_from_dtt[:, 3]
    coh_E1 = acc.coh('H1:LSC-DARM1_EXC',  'H1:LSC-DARM1_IN1')
    coh_E2 = acc.coh('H1:LSC-DARM1_EXC',  'H1:LSC-DARM1_IN2')
    coh_12 = acc.coh('H1:LSC-DARM1_IN1',  'H1:LSC-DARM1_IN2')
    assert_allclose(freq_from_dtt, coh_E1.FHz)
    assert_allclose(COH_E1_from_dtt, coh_E1.coh)
    assert_allclose(COH_E2_from_dtt, coh_E2.coh)
    assert_allclose(COH_12_from_dtt, coh_12.coh)

    #now do CSD
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB_CSD_E1_E2_12.txt'),
        dtype='float',
        delimiter=None
    )
    freq_from_dtt = data_from_dtt[:, 0]
    CSD_E1_from_dtt = data_from_dtt[:, 1] + 1j*data_from_dtt[:, 2]
    CSD_E2_from_dtt = data_from_dtt[:, 3] + 1j*data_from_dtt[:, 4]
    CSD_12_from_dtt = data_from_dtt[:, 5] + 1j*data_from_dtt[:, 6]
    csd_E1 = acc.csd('H1:LSC-DARM1_EXC',  'H1:LSC-DARM1_IN1')
    csd_E2 = acc.csd('H1:LSC-DARM1_EXC',  'H1:LSC-DARM1_IN2')
    csd_12 = acc.csd('H1:LSC-DARM1_IN1',  'H1:LSC-DARM1_IN2')
    assert_allclose(freq_from_dtt, csd_E1.FHz)
    assert_allclose(CSD_E1_from_dtt, csd_E1.csd)
    assert_allclose(CSD_E2_from_dtt, csd_E2.csd)
    assert_allclose(CSD_12_from_dtt, csd_12.csd)





import unittest
import numpy as np
from numpy.testing import assert_allclose

import pydarm_measurement as measurement


def test_get_raw_tf(fpath_join, tpath_join):
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2020-01-03_H1_'
                   'DARM_OLGTF_LF_SS_A_DARMIN2_B_DARMIN1_tf.txt'),
        dtype='float',
        delimiter=None
    )

    freq_from_dtt = data_from_dtt[:, 0]
    mag_from_dtt = np.abs(data_from_dtt[:, 1]+1j*data_from_dtt[:, 2])
    angle_from_dtt = np.angle(
        data_from_dtt[:, 1] + 1j*data_from_dtt[:, 2],
        deg=True
    )

    measurement_file = fpath_join('data', '2020-01-03_H1_'
                                  'DARM_OLGTF_LF_SS_5to1100Hz_15min.xml')
    channelA = 'H1:LSC-DARM1_IN2'
    channelB = 'H1:LSC-DARM1_IN1'
    meas_object = measurement.Measurement(measurement_file)
    freq, tf, coh, unc = meas_object.get_raw_tf(channelA, channelB)

    # Note: For some reason, the txt file is missing one frequency.
    # This is unrelated to the coherence, at least of this TF.
    freq_from_xml = freq
    freq_from_xml = np.delete(freq_from_xml, 22)
    mag_from_xml = np.abs(tf)
    mag_from_xml = np.delete(mag_from_xml, 22)
    angle_from_xml = np.angle(tf, deg=True)
    angle_from_xml = np.delete(angle_from_xml, 22)

    assert_allclose(freq_from_dtt, freq_from_xml, atol=1e-4)
    assert_allclose(mag_from_dtt, mag_from_xml, atol=1e-4)
    assert_allclose(angle_from_dtt, angle_from_xml, atol=1e-4)


def test_get_raw_asd(fpath_join, pprint):
    data_from_dtt = np.genfromtxt(
        fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.txt'),
        dtype='float',
        delimiter=None
    )

    freq_from_dtt = data_from_dtt[:, 0]
    asd_from_dtt = data_from_dtt[:, 1]

    measurement_file = fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.xml')
    channelA = 'H1:LSC-DARM1_IN2'
    meas_object = measurement.Measurement(measurement_file)
    freq, asd = meas_object.get_raw_asd(channelA)

    freq_from_xml = freq
    asd_from_xml = asd

    pprint(asd_from_dtt, asd_from_xml)
    assert_allclose(freq_from_dtt, freq_from_xml)
    assert_allclose(asd_from_dtt, asd_from_xml)


def test_get_set_of_channels(fpath_join):
    channels_to_compare = [
        'H1:LSC-DARM1_EXC',
        'H1:LSC-DARM1_IN1',
        'H1:LSC-DARM1_IN2']

    measurement_file = fpath_join('data', '2019-03-27_H1DARM_OLGTF_BB.xml')
    meas_object = measurement.Measurement(measurement_file)
    set_of_channels = meas_object.get_set_of_channels()

    assert(not (set(set_of_channels) ^ set(channels_to_compare)))



"""
"""
from __future__ import print_function, division

import warnings
import numpy as np

from declarative.bunch.deep_bunch import DeepBunch

from .access import DiagAccess


class DiagFileError(ValueError):
    pass


def dtt2bunch(
    fpath,
    channels         = None,
    no_remap         = True,
    channel_map      = {},
    channels_exclude = set(),
    verbose          = False,
):
    try:
        daccess = DiagAccess(fpath)
    except Exception:
        raise DiagFileError("Diag File malformed (xml syntax error)")
    def insert_refs(dbunch):
        for k, v in daccess.references.items():
            dbunch.REFS[str(k)] = v

    Achns, Bchns = daccess.channels()
    allchns_raw = set(Achns)
    allchns_raw.update(Bchns)
    allchns_raw = list(allchns_raw)
    allchns_raw.sort()

    spectra_bunch = DeepBunch()
    spectra_bunch.type = 'spectra'
    #PSD type
    xsd_current = None
    xsd_prev    = None
    xfer_current = None
    xfer_prev    = None
    for idx_A in range(len(allchns_raw)):
        chn_A_raw = allchns_raw[idx_A]
        if chn_A_raw in channels_exclude:
            continue
        try:
            chn_A = channel_map[chn_A_raw]
        except KeyError:
            if no_remap:
                chn_A = chn_A_raw
            else:
                continue
        if channels is not None and chn_A not in channels:
            continue
        if chn_A in channels_exclude:
            continue
        for idx_B in range(len(allchns_raw)):
            chn_B_raw = allchns_raw[idx_B]
            if chn_B_raw in channels_exclude:
                continue
            try:
                chn_B = channel_map[chn_B_raw]
            except KeyError:
                if no_remap:
                    chn_B = chn_B_raw
                else:
                    continue
            if channels is not None and chn_B not in channels:
                continue

            if chn_B in channels_exclude:
                continue

            if verbose:
                print(chn_A, chn_B)
            if chn_A == chn_B:
                try:
                    xsd_current = daccess.asd(chn_A_raw)
                    value = xsd_current.asd
                    prev = spectra_bunch.ASD.setdefault(chn_A, value)
                    if prev is not value:
                        if not all(prev == value):
                            warnings.warn((
                                "Multiple Mappings from raw to real channel are not consistent for ASD of chn: {0}"
                            ).format(chn_A)
                            )
                except KeyError:
                    continue
            else:
                try:
                    chn_A_dict = spectra_bunch.CSD.setdefault(chn_A, {})
                    xsd_current = daccess.csd(chn_A_raw, chn_B_raw)
                    value = xsd_current.csd
                    prev = chn_A_dict.setdefault(chn_B, value)
                    if prev is not value:
                        if not all(prev == value):
                            warnings.warn((
                                "Multiple Mappings from raw to real channel are not consistent for CSD between chns: {0} and {1}"
                            ).format(chn_A, chn_B)
                            )
                    if verbose:
                        print("GOOD: ", chn_A, chn_B)
                except KeyError:
                    pass

                try:
                    chn_A_dict = spectra_bunch.XFER.setdefault(chn_A, {})
                    xfer_current = daccess.xfer(chn_A_raw, chn_B_raw)
                    value = xfer_current.xfer
                    prev = chn_A_dict.setdefault(chn_B, value)
                    if prev is not value:
                        if not all(prev == value):
                            warnings.warn((
                                "Multiple Mappings from raw to real channel are not consistent for XFER between chns: {0} and {1}"
                            ).format(chn_A, chn_B)
                            )
                except KeyError:
                    print("BAD: ", chn_A, chn_B)
                    continue

                chn_A_dict = spectra_bunch.COH.setdefault(chn_A, {})
                value = daccess.coh(chn_A_raw, chn_B_raw).coh
                prev = chn_A_dict.setdefault(chn_B, value)
                if prev is not value:
                    if not all(prev == value):
                        warnings.warn((
                            "Multiple Mappings from raw to real channel are not consistent for COH between chns: {0} and {1}"
                        ).format(chn_A, chn_B)
                        )
                chn_A_dict = spectra_bunch.XFER_SNR_EST.setdefault(chn_A, {})
                value = daccess.xfer(chn_A_raw, chn_B_raw).SNR_estimate
                value = np.maximum(value**2 - 1, 0)
                prev = chn_A_dict.setdefault(chn_B, value)
                if prev is not value:
                    if not all(prev == value):
                        warnings.warn((
                            "Multiple Mappings from raw to real channel are not consistent for XFER_SNR_EST between chns: {0} and {1}"
                        ).format(chn_A, chn_B)
                        )
            if xsd_prev is not None:
                xsd_prev.metadata_check(xsd_current)
            xsd_prev = xsd_current

            if xfer_prev is not None:
                xfer_prev.metadata_check(xfer_current)
            xfer_prev = xfer_current

    if xsd_current is not None:
        spectra_bunch.gps_second   = xsd_current.gps_second
        spectra_bunch.window       = xsd_current.window
        spectra_bunch.averages     = xsd_current.averages
        spectra_bunch.BW           = xsd_current.BW
        spectra_bunch.FHz          = xsd_current.FHz.squeeze()
        insert_refs(spectra_bunch)
        return spectra_bunch
    elif xfer_current is not None:
        spectra_bunch.gps_second   = xfer_current.gps_second
        #spectra_bunch.window       = xfer_current.window
        spectra_bunch.averages     = xfer_current.averages
        #spectra_bunch.BW           = xfer_current.BW
        spectra_bunch.FHz          = xfer_current.FHz.squeeze()
        insert_refs(spectra_bunch)
        return spectra_bunch

    TS_bunch = DeepBunch()
    TS_bunch.type = 'timeseries'
    try:
        diag_TS = daccess.results.TS
    except (KeyError, AttributeError):
        raise DiagFileError("No Channels Found in TS")
    ts_current = None
    ts_prev = None
    for idx_A in range(len(allchns_raw)):
        chn_A_raw = allchns_raw[idx_A]
        if chn_A_raw in channels_exclude:
            continue
        try:
            chn_A = channel_map[chn_A_raw]
        except KeyError:
            if no_remap:
                chn_A = chn_A_raw
            else:
                continue
        if channels is not None and chn_A not in channels:
            continue
        if chn_A in channels_exclude:
            continue
        try:
            ts_current = diag_TS[chn_A_raw]
        except KeyError:
            pass
        else:
            value = ts_current.timeseries
            prev = TS_bunch.TS.setdefault(chn_A, value)
            ts_prev = ts_current
            if prev is not value:
                if not all(prev == value):
                    warnings.warn((
                        "Multiple Mappings from raw to real channel are not consistent for TS of chn: {0}"
                    ).format(chn_A)
                    )
    if ts_current is not None:
        TS_bunch.gps_second    = ts_current.gps_second
        TS_bunch.time_delay_s  = ts_current.time_delay_s
        TS_bunch.avgtype       = ts_current.avgtype
        TS_bunch.dt            = ts_current.dt
        insert_refs(TS_bunch)
        return TS_bunch
    else:
        raise DiagFileError("No Channels Found in TS")





"""
"""

import base64
import numpy as np
from declarative.bunch import Bunch
import re


def parse_coefficients(
    LW_node,
    entry_type = None,
    entry_num = None,
    entry_text = None
):
    if entry_type == 'TransferMatrix':
        return None
    specbunch = Bunch()
    specbunch.channelA = {}
    specbunch.channelB = {}
    specbunch.channels = {}

    for sparam in LW_node.findall('Time'):
        if sparam.attrib["Name"] == 't0':
            specbunch.gps_second = float(sparam.text)
    for sparam in LW_node.findall('Param'):
        spname = sparam.attrib["Name"]
        if spname == 'Subtype':
            specbunch.subtype_raw = int(sparam.text)
        elif spname == 'Window':
            specbunch.window_raw = int(sparam.text)
            specbunch.window = {
                0 : 'uniform',
                1 : 'hanning',
                2 : 'flat-top',
                3 : 'welch',
                4 : 'bartlett',
                5 : 'bmh',
                6 : 'hamming',
                7 : 'kaiser',
            }.get(specbunch.window_raw, "unknown")
        elif spname == 'AverageType':
            specbunch.avgtype_raw = int(sparam.text)
            specbunch.avgtype = {
                0 : 'Fixed',
                1 : 'Exponential',
                2 : 'Single',
            }.get(specbunch.avgtype_raw, "unknown")
        elif spname == 'Averages':
            specbunch.averages = int(sparam.text)
        elif spname == 'BW':
            specbunch.BW = float(sparam.text)
        elif spname == 'M':
            specbunch.M = int(sparam.text)
        elif spname == 'N':
            specbunch.N = int(sparam.text)
        elif spname == 'f0':
            specbunch.f0 = float(sparam.text)
        elif spname == 'df':
            specbunch.df = float(sparam.text)
        elif spname.startswith('ChannelA'):
            RefNMatch = re.match(r'ChannelA\[(\d+)\]', spname)
            chn_num = int(RefNMatch.group(1))
            specbunch.channelA[chn_num] = sparam.text
            specbunch.channels[chn_num] = sparam.text
        elif spname.startswith('ChannelB'):
            RefNMatch = re.match(r'ChannelB\[(\d+)\]', spname)
            chn_num = int(RefNMatch.group(1))
            specbunch.channelB[chn_num] = sparam.text
            specbunch.channels[chn_num] = sparam.text
        else:
            pass

    chn_dict = specbunch.setdefault('channelA', {})
    specbunch.channelA_inv = {}
    for idx, chn_name in chn_dict.items():
        specbunch.channelA_inv[chn_name] = idx

    chn_dict = specbunch.setdefault('channelB', {})
    specbunch.channelB_inv = {}
    for idx, chn_name in chn_dict.items():
        specbunch.channelB_inv[chn_name] = idx

    chn_dict = specbunch.setdefault('channels', {})
    specbunch.channels_inv = {}
    for idx, chn_name in chn_dict.items():
        specbunch.channels_inv[chn_name] = idx

    ####
    ##Now interpret the data
    array_node = LW_node.findall('Array')[0]
    dims = [int(d.text) for d in array_node.findall('Dim')]
    streambuff = base64.b64decode(array_node.findall('Stream')[0].text)
    stream_type = array_node.attrib['Type']
    if stream_type == 'float':
        data = np.frombuffer(streambuff, dtype='f4')
    elif stream_type == 'floatComplex':
        data = np.frombuffer(streambuff, dtype='c8')
    data = data.reshape(*dims)

    specbunch.data_raw = data
    specbunch.FHz = np.real(data[:, 0])
    specbunch.coeffs = data[:, 1:]

    if entry_type == 'CoherenceCoefficients':
        specbunch.coeffs = np.real(specbunch.coeffs)

    #specbunch.FHz = data[:N].real
    if specbunch.subtype_raw == 0:
        specbunch.subtype = "1-D fixed bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 1:
        specbunch.subtype = "1-D variable bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 2:
        specbunch.subtype = "2-D fixed bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 3:
        specbunch.subtype = "2-D variable bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 4:
        specbunch.subtype = "3-D fixed bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 5:
        specbunch.subtype = "3-D varibale bin spacing histogram"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 6:
        specbunch.subtype = "1-D fixed bin spacing histogram with errors"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 7:
        specbunch.subtype = "1-D variable bin spacing histogram with errors"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 8:
        specbunch.subtype = "2-D fixed bin spacing histogram with errors"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 9:
        specbunch.subtype = "2-D variable bin spacing histogram with errors"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 10:
        specbunch.subtype = "3-D fixed bin spacing histogram with errors"
        specbunch.type_name = entry_type
    elif specbunch.subtype_raw == 11:
        specbunch.subtype = "3-D variable bin spacing histogram with errors"
        specbunch.type_name = entry_type
    else:
        specbunch.subtype = "unknown"
    return specbunch





import numpy as np
import dttxml
from os import path

class Measurement(object):
    """
    This class is used to get measurement data from
    a dtt measurement. This class is used in conjuction with the
    ProcessSensingMeasurement and ProcessActuationMeasurement
    class to get the input for the MCMC and GPR analysis.

    This class uses the dttxml module to import transfer functions and ASDs.

    Parameters
    ----------
    meas_file: string
        String of the name of the xml file of the measurement.

    Returns
    -------

    Methods (Quick reference)
    -------
    get_raw_tf: Input two channels, get array of freq and TFs
    get_raw_asd: Input one channel, get array of freq and ASD
    get_set_of_channels: No input, get a set of the channels in meas_file.

    NOTES
    -----
    If the file is not given, the class will complain.
    """

    def __init__(self, meas_file):
        if path.exists(meas_file) is not True:
            raise IOError('The dtt file given does not exist.')

        # self.measurement = "sensing"

        # Data accessing object. It is called later.
        self.data_access = dttxml.DiagAccess(meas_file)

        # FIXME
        # For now, I can only access the averages and gps time from the
        # file only after I get the transfer function. There must be
        # a way to get them just after we access the file. For now,
        # I just initialize them here.
        self.averages = None
        self.gps_time = None

    def get_raw_tf(self, channelA, channelB, cohThresh=0):
        """
        Given two channels it returns the transfer function from the dtt
        file. Note that it gives channelB/channelA.
        In addition it populates averages and gps_time, calculates the
        uncertainty from coherence, and rejects data with coherence lower
        than the given threshold.

        Parameters
        ----------
        channelA: string
            Channel in the denominator of the TF
        channelB: string
            Channel in the numerator of the TF
        cohThresh: float
            Coherence threshold below which data from dtt are rejected

        Returns
        -------
        Returns a comlex array of (Nx4), where N is the number of data
        points, and the four columns are:
        Column 1: frequencies (frequencies of the data in Hz)
        Column 2: tf (transfer function of B/A)
        Column 3: coh (coherence of B/A measurement)
        Column 4: unc (the uncertainty from coherence)
        """
        if (channelA in self.get_set_of_channels()) and\
           (channelB in self.get_set_of_channels()):
            # Transfer function data holder. xfer is a method that comes from
            # dttxml. xfer takes channelB first and channelA second, which
            # is the opposite order that I am using. Whatever the ordering,
            # channelA always means the denominator and channelB is the
            # numerator.This should work for both Swept Sine
            # or FFT (aka broadband) measurements.
            tf_holder = self.data_access.xfer(channelB, channelA)
            self.averages = tf_holder.averages
            self.gps_time = tf_holder.gps_second
            # Frequency of the measurement in Hz
            frequencies = np.array(tf_holder.FHz)
            # Transfer function of B/A, complex array
            tf = np.array(tf_holder.xfer)
            # Coherence function of B/A, real array
            coh = np.array(tf_holder.coh)
            # Uncertainty from coherence
            unc = np.sqrt(1.0-coh) / \
                (2.0*(coh+1e-6)*self.averages)
            # Find good coherence and intersecting points
            frequencies = frequencies[(coh > cohThresh)]
            tf = tf[(coh > cohThresh)]
            unc = unc[(coh > cohThresh)]
            coh = coh[(coh > cohThresh)]

            return frequencies, tf, coh, unc
        else:
            raise ValueError('Invalid channel for transfer function.')

    def get_raw_asd(self, channelA):
        """
        It returns the ASD of channelA from the dtt file.
        This will only work for a FFT (aka broadband) type measurement.

        Parameters
        ----------
        channelA: string
            ASD channel.

        Returns
        -------
        Returns two float array:
        Array 1: frequencies (frequencies of the data in Hz)
        Array 2: asd (ASD of channelA)
        """

        if (channelA in self.get_set_of_channels()):
            # This will only work for a FFT (aka broadband) type measurement
            asd_holder = self.data_access.asd(channelA)
            self.averages = asd_holder.averages
            self.gps_time = asd_holder.gps_second
            # Frequency of the measurement in Hz
            frequencies = np.array(asd_holder.FHz)
            # Amplitude Spectral Density of requested channel
            asd = np.array(asd_holder.asd)
            return frequencies, asd
        else:
            raise ValueError('Invalid channel for asd.')

    def get_set_of_channels(self):
        """
        Method to get the channels in the measurement file.
        Annoyingly, the channels are given in a set.
        Parameters
        ----------

        Returns
        -------
        channels: A python set with all the channels in the measurement file
        of this object.
        """
        channels = self.data_access.channels()
        return channels[0]



"""
"""

import base64
import numpy as np
from declarative.bunch import Bunch


def parse_timeseries(
    LW_node,
):
    timebunch = Bunch()

    for sparam in LW_node.findall('Time'):
        if sparam.attrib["Name"] == 't0':
            timebunch.gps_second = float(sparam.text)
    for sparam in LW_node.findall('Param'):
        spname = sparam.attrib["Name"]
        if spname == 'Subtype':
            timebunch.subtype_raw = int(sparam.text)
        elif spname == 'AverageType':
            timebunch.avgtype_raw = int(sparam.text)
            timebunch.avgtype = {
                0 : 'Fixed',
                1 : 'Exponential',
                2 : 'Single',
            }.get(timebunch.avgtype_raw, "unknown")
        elif spname == 'Averages':
            timebunch.averages = int(sparam.text)
        elif spname == 'N':
            timebunch.N = int(sparam.text)
        elif spname == 'f0':
            timebunch.f0 = float(sparam.text)
        elif spname == 'dt':
            timebunch.dt = float(sparam.text)
        elif spname == 'Channel':
            timebunch.channel = sparam.text
        elif spname == "Decimation1":
            timebunch.decimation1 = int(sparam.text)
        elif spname == "DecimationType":
            timebunch.decimation_rawtype = int(sparam.text)
        elif spname == "DecimationDelay":
            timebunch.decimation_delay_s = float(sparam.text)
        elif spname == "TimeDelay":
            timebunch.time_delay_s = float(sparam.text)
        elif spname == "DelayTaps":
            timebunch.delay_taps_num = int(sparam.text)
        elif spname == "DecimationFilter":
            timebunch.decimation_filter = sparam.text
        else:
            pass

    ####
    ##Now interpret the data
    array_node = LW_node.findall('Array')[0]
    dims = [int(d.text) for d in array_node.findall('Dim')]
    streambuff = base64.b64decode(array_node.findall('Stream')[0].text)
    stream_type = array_node.attrib['Type']
    if stream_type == 'float':
        data = np.frombuffer(streambuff, dtype='f4')
    elif stream_type == 'floatComplex':
        data = np.frombuffer(streambuff, dtype='c8')
    data = data.reshape(*dims)

    timebunch.data_raw = data

    timebunch.type_name = 'TS'
    if timebunch.subtype_raw == 0:
        timebunch.subtype = "normal time series in format (Y)"
        timebunch.timeseries = data
    elif timebunch.subtype_raw == 1:
        timebunch.subtype = "down-converted time series in format (Y)"
        timebunch.timeseries = data
    elif timebunch.subtype_raw == 2:
        timebunch.subtype = "averaged time series in format (Y)"
        timebunch.timeseries = data
    elif timebunch.subtype_raw == 3:
        timebunch.subtype = "averaged time series in format (mean, std. dev., min., max., rms)"
    elif timebunch.subtype_raw == 4:
        timebunch.subtype = "normal time series in format (t,Y)"
    elif timebunch.subtype_raw == 5:
        timebunch.subtype = "down-converted time series in format (t,Y)"
    elif timebunch.subtype_raw == 6:
        timebunch.subtype = "averaged time series in format (t,Y)"
    elif timebunch.subtype_raw == 7:
        timebunch.subtype = "averaged time series in format (t, mean, std. dev., min., max., rms)"
    else:
        timebunch.subtype = "unknown"
        timebunch.type_name = "unknown"
    return timebunch





"""
"""
import xml.etree.cElementTree as etree
import numpy as np
import re

from declarative.bunch import Bunch
from .parse_transfer import parse_transfer
from .parse_spectrum import parse_spectrum
from .parse_timeseries import parse_timeseries
from .parse_coefficients import parse_coefficients


def coherence_TF_numerator_SNR(coherence, N_averages = None, rejection_ratio = 1.5):
    #limit it to about SNR of 2000 for numerical accuracy reasons
    #coherence = np.minimum(coherence, .9999999)
    coherence = np.minimum(coherence, .99999)
    single_snr = coherence**2 / (1 - coherence**2)
    if N_averages is None:
        return single_snr
    minimum_coh = 1/N_averages**.5
    snr = single_snr * N_averages**.5
    snr[coherence < minimum_coh * rejection_ratio ] = 0
    return snr


def entry_type_num(text):
    elines = list(text.splitlines())
    elines[0]
    match_ref = re.search(r'(\w+)\[(\d+)\].*', elines[0])
    if match_ref:
        return match_ref.group(1), match_ref.group(2)
    match_ref = re.search(r'(\w+).*', elines[0])
    if match_ref:
        return match_ref.group(1), None
    return elines[0]


def dtt_read(diag_file):
    references = {}
    results = {}

    diag_tree = etree.parse(diag_file)
    entries = {}

    for xml_node in diag_tree.getroot():
        if xml_node.attrib['Name'] == 'Index':
            for subentry in xml_node:
                match_ref = re.match(r'Entry\[(\d+)\]', subentry.attrib['Name'])
                if match_ref:
                    ref_num = int(match_ref.group(1))
                    entries[ref_num] = subentry.text

    for entry_text in entries.values():
        entry_type, entry_num = entry_type_num(entry_text)
        if entry_type in (
            'TransferCoefficients',
            'HarmonicCoefficients',
            'IntermodulationCoefficients',
            'CoherenceCoefficients',
            #'TransferMatrix,
        ):
            entry_parse = re.search(r'Name\s+=\s+(\w+.*);', entry_text)
            entry_node_name = entry_parse.group(1)
            for xml_node in diag_tree.getroot().findall('LIGO_LW'):
                if xml_node.attrib['Name'] == entry_node_name:
                    break
            else:
                print("WARNING XML is funky!")
            measurement = parse_coefficients(
                LW_node = xml_node,
                entry_type = entry_type,
                entry_num = entry_num,
                entry_text = entry_text,
            )
            old = results.setdefault(entry_type, measurement)
            assert(old == measurement)

    for xml_node in diag_tree.getroot():
        match_ref = re.match(r'Reference\[(\d+)\]', xml_node.attrib['Name'])
        if match_ref:
            ref_num = int(match_ref.group(1))
            typename = xml_node.attrib['Type']
            if typename == "TransferFunction":
                ref = parse_transfer(xml_node)
                references[ref_num] = ref
            elif typename == "Spectrum":
                ref = parse_spectrum(xml_node)
                references[ref_num] = ref
            elif typename == "TimeSeries":
                ref = parse_timeseries(xml_node)
                references[ref_num] = ref
        match_ref = re.match(r'Result\[(\d+)\]', xml_node.attrib['Name'])
        if match_ref:
            ref_num = int(match_ref.group(1))
            typename = xml_node.attrib['Type']
            ref = None
            if typename == "TransferFunction":
                ref = parse_transfer(xml_node)
            elif typename == "Spectrum":
                ref = parse_spectrum(xml_node)
            elif typename == "TimeSeries":
                ref = parse_timeseries(xml_node)
            if ref is not None:
                try:
                    if ref.type_name in ('COH', 'STF', 'TF', 'PSD', 'FFT', 'CSD'):
                        results.setdefault(ref.type_name, {})[ref.channelA] = ref
                    elif ref.type_name in ('TS',):
                        results.setdefault(ref.type_name, {})[ref.channel] = ref
                    else:
                        #Check that these routines aren't returning something unaccounted-for
                        assert(False)
                except KeyError:
                    continue

    Items = Bunch()
    Items.references = Bunch(references)
    Items.results = Bunch(results)
    return Items





import os
import setup_helper
from setuptools import setup

with open(os.path.join('dttxml', 'version.py')) as f:
    exec(f.read())
cmdclass = setup_helper.version_checker(version, 'dttxml')


setup(
    name='dttxml',
    version=version,
    url='https://git.ligo.org/cds/dttxml',
    author='Lee McCuller',
    author_email='Lee.McCuller@ligo.org',
    description=(
        'Extract data from LIGO Diagnostics test tools XML format. Formerly dtt2hdf.'
    ),
    license='Apache v2',
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    packages=[
        'dttxml',
    ],
    install_requires=[
        'numpy', 'h5py',
        'declarative>=1.3.0',
    ],
    extras_require={
        "hdf" : ["h5py"],
    },
    tests_require=[
        'pytest',
    ],
    entry_points={
        'console_scripts': [
            'dtt2hdf=dttxml.dtt2hdf:main',
        ],
    },
    cmdclass=cmdclass,
    zip_safe=True,
    keywords='LIGO dtt diagnostics file-reader',
)



# -*- coding: utf-8 -*-
"""
"""
from .parse import dtt_read
from .access import DiagAccess
from .dtt2bunch import dtt2bunch

from .version import (
    version,
    __version__,
)

__all__ = [
    dtt_read,
    DiagAccess,
    dtt2bunch,
    version,
    __version__,
]
